#ifndef _IP_NAT_H
#define _IP_NAT_H
#include <linux/netfilter_ipv4.h>
#include <linux/netfilter_ipv4/ip_conntrack_tuple.h>

#define IP_NAT_MAPPING_TYPE_MAX_NAMELEN 16

enum ip_nat_manip_type
{
	IP_NAT_MANIP_SRC,
	IP_NAT_MANIP_DST
};

/* SRC manip occurs POST_ROUTING or LOCAL_IN */
#define HOOK2MANIP(hooknum) ((hooknum) != NF_IP_POST_ROUTING && (hooknum) != NF_IP_LOCAL_IN)

#define IP_NAT_RANGE_MAP_IPS 1
#define IP_NAT_RANGE_PROTO_SPECIFIED 2
#define IP_NAT_RANGE_BYPASS_HELPERS 4

#ifdef CONFIG_IP_NF_NAT_NRES
#define IP_NAT_RANGE_USE_RESERVED 8
#endif

/* NAT sequence number modifications */
struct ip_nat_seq {
	/* position of the last TCP sequence number 
	 * modification (if any) */
	u_int32_t correction_pos;
	/* sequence number offset before and after last modification */
	int16_t offset_before, offset_after;
};

/* Single range specification. */
struct ip_nat_range
{
	/* Set to OR of flags above. */
	unsigned int flags;

	/* Inclusive: network order. */
	__be32 min_ip, max_ip;

	/* Inclusive: network order */
	union ip_conntrack_manip_proto min, max;
};

/* For backwards compat: don't use in modern code. */
struct ip_nat_multi_range_compat
{
	unsigned int rangesize; /* Must be 1. */

	/* hangs off end. */
	struct ip_nat_range range[1];
};

#ifdef __KERNEL__
#include <linux/list.h>

/* call this to signal dependency on the NAT hooks */
extern void need_nat_hooks(void);

/* Protects NAT hash tables, and NAT-private part of conntracks. */
extern rwlock_t ip_nat_lock;

/* The structure embedded in the conntrack structure. */
struct ip_nat_info
{
	struct list_head bysource;
	struct ip_nat_seq seq[IP_CT_DIR_MAX];
};

struct ip_conntrack;

#ifdef CONFIG_IP_NF_NAT_NRES
/* Structure to store reserved manips */
struct ip_nat_reserved {
	struct list_head hash;			/* Hash chain */
	struct list_head exp;			/* Per-expectation list */
	atomic_t use;				/* Reference count */
	struct ip_conntrack_manip manip;	/* Reserved manip */
	struct ip_conntrack_manip peer;		/* Peer (optional) */
	u_int16_t proto;			/* Protocol number of reserved manip */
};
#endif

/* Set up the info structure to map into this range. */
extern unsigned int ip_nat_setup_info(struct ip_conntrack *conntrack,
				      const struct ip_nat_range *range,
				      unsigned int hooknum);

/* Is this tuple already taken? (not by us)*/
extern int ip_nat_used_tuple(const struct ip_conntrack_tuple *tuple,
			     const struct ip_conntrack *ignored_conntrack,
			     const enum ip_nat_manip_type maniptype,
			     const unsigned int flags);

#ifdef CONFIG_IP_NF_NAT_NRES
struct ip_conntrack_expect;

/* NAT port reservation: allocate and hash a new entry */
extern struct ip_nat_reserved *__ip_nat_reserved_new_hash(const struct ip_conntrack_manip *manip,
					 const u_int16_t proto, const struct ip_conntrack_manip *peer);

/* NAT port reservation: unhash an entry */
extern struct ip_nat_reserved *__ip_nat_reserved_unhash(const struct ip_conntrack_manip *manip,
				       const u_int16_t proto, const struct ip_conntrack_manip *peer);

/* NAT port reservation: free a reservation */
extern void __ip_nat_reserved_free(struct ip_nat_reserved *res);

/* NAT port reservation: register a new reservation */
extern int ip_nat_reserved_register(struct ip_conntrack_expect *exp,
				    const struct ip_conntrack_manip *manip,
				    const u_int16_t proto,
				    const struct ip_conntrack_manip *peer);

/* NAT port reservation: unregister a reservation */
extern int ip_nat_reserved_unregister(struct ip_conntrack_expect *exp,
				      const struct ip_conntrack_manip *manip,
				      const u_int16_t proto,
				      const struct ip_conntrack_manip *peer);

/* NAT port reservation: unregister all reservations for a given expectation */
extern void ip_nat_reserved_unregister_all(struct ip_conntrack_expect *exp);
#endif /*CONFIG_IP_NF_NAT_NRES*/

#else  /* !__KERNEL__: iptables wants this to compile. */
#define ip_nat_multi_range ip_nat_multi_range_compat
#endif /*__KERNEL__*/
#endif
