/*
 * Transparent proxy support for Linux/iptables
 *
 * Copyright (c) 2002-2004 BalaBit IT Ltd.
 * Author: Balzs Scheidler
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/ip.h>
#include <linux/inetdevice.h>
#include <net/checksum.h>

#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ipt_TPROXY.h>
#include <linux/netfilter_ipv4/iptable_tproxy.h>

/* determine ip address of the interface the packet came in */
static u32
determine_local_ip(struct sk_buff *skb, int hooknum)
{
	struct in_device *indev;
	u32 ip;

	if (hooknum == NF_IP_LOCAL_OUT)
		return htonl(0x7f000001);

	indev = in_dev_get(skb->dev);

	if (!indev) {
		printk(KERN_WARNING "IP_TPROXY: No IP protocol on incoming "
		       "interface during redirect, dropping packet.\n");
		return 0;
	}
	if (!indev->ifa_list) {
		printk(KERN_WARNING "IP_TPROXY: No IP address on incoming "
		       "interface during redirect, dropping packet.\n");
		in_dev_put(indev);
		return 0;
	}

	ip = indev->ifa_list->ifa_local;
	in_dev_put(indev);

	return ip;
}

static unsigned int
target(struct sk_buff **pskb,
       const struct net_device *in,
       const struct net_device *out,
       unsigned int hooknum,
       const struct xt_target *target,
       const void *targinfo)
{
	const struct ipt_tproxy_target_info *info = targinfo;
	struct ip_tproxy_sockref sr;
	u16 *ports, _ports[2];
	struct iphdr *iph = (*pskb)->nh.iph;

	/* TCP/UDP only */
	if ((iph->protocol != IPPROTO_TCP) &&
	    (iph->protocol != IPPROTO_UDP))
		return NF_ACCEPT;

	/* get ports */
	ports = skb_header_pointer(*pskb, iph->ihl * 4,
				   sizeof(_ports), &_ports);
	if (ports == NULL)
		return NF_DROP;

	/* set up fake sockref */
	memset(&sr, 0, sizeof(sr));
	sr.proto = iph->protocol;

	/* use the original address/port if none is specified in the rule */
	if (info->laddr)
		sr.laddr = info->laddr;
	else
		sr.laddr = determine_local_ip(*pskb, hooknum);

	if (info->lport)
		sr.lport = info->lport;
	else
		sr.lport = ports[1];

	/* set up mapping */
	return ip_tproxy_setup_nat(pskb, hooknum, &sr, 0);
}

static int
checkentry(const char *tablename,
	   const void *e,
	   const struct xt_target *target,
           void *targinfo,
           unsigned int hook_mask)
{
	if (strcmp(tablename, "tproxy") != 0) {
		printk(KERN_WARNING "TPROXY: can only be called from \"tproxy\" table, not \"%s\"\n", tablename);
		return 0;
	}

	return 1;
}

static struct ipt_target ipt_tproxy_reg = {
	.name		= "TPROXY",
	.target		= target,
	.targetsize	= sizeof(struct ipt_tproxy_target_info),
	.checkentry	= checkentry,
	.me		= THIS_MODULE,
};

static int __init init(void)
{
	if (ipt_register_target(&ipt_tproxy_reg))
		return -EINVAL;

	return 0;
}

static void __exit fini(void)
{
	ipt_unregister_target(&ipt_tproxy_reg);
}

module_init(init);
module_exit(fini);
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Balzs Scheidler <bazsi@balabit.hu>");
MODULE_DESCRIPTION("Netfilter transparent proxy TPROXY target module.");
